@@ -49,6 +49,8 @@ typedef struct {
   gchar *relay_token;
 } TfNatProperties;
 
+typedef void (NewStreamCreatedCb) (TfStream *stream, gpointer channel);
+
 TfStream *
 _tf_stream_new (gpointer channel,
     FsConference *conference,
@@ -59,13 +61,15 @@ _tf_stream_new (gpointer channel,
     TpMediaStreamDirection direction,
     TfNatProperties *nat_props,
     GList *local_codecs_config,
-    GError **error);
+    NewStreamCreatedCb new_stream_created_cb);
 
 gboolean _tf_stream_bus_message (TfStream *stream,
     GstMessage *message);
 
 void _tf_stream_try_sending_codecs (TfStream *stream);
 
+TpMediaStreamError fserror_to_tperror (GError *error);
+
 G_END_DECLS
 
 #endif /* __TF_STREAM_PRIV_H__ */
