@@ -67,6 +67,11 @@ G_DEFINE_TYPE (TfStream, tf_stream, G_TYPE_OBJECT);
 
 #define STREAM_PRIVATE(o) ((o)->priv)
 
+#define TF_STREAM_LOCK(o)   (g_static_mutex_lock (&(o)->priv->mutex))
+#define TF_STREAM_UNLOCK(o) (g_static_mutex_unlock (&(o)->priv->mutex))
+
+static TpMediaStreamError fserrorno_to_tperrorno (FsError fserror);
+
 struct _TfStreamPrivate
 {
   TfChannel *channel;
@@ -94,6 +99,12 @@ struct _TfStreamPrivate
 
   guint tos;
 
+  GStaticMutex mutex;
+  guint idle_connected_id; /* Protected by mutex */
+  gboolean disposed; /* Protected by mutex */
+
+  TpMediaStreamState current_state;
+
   NewStreamCreatedCb *new_stream_created_cb;
 };
 
@@ -190,6 +201,11 @@ static void cb_fs_stream_src_pad_added (FsStream *fsstream G_GNUC_UNUSED,
     FsCodec *codec,
     gpointer user_data);
 
+static void cb_fs_component_state_changed (TfStream *self,
+    guint component,
+    FsStreamState fsstate);
+
+
 static void
 tf_stream_init (TfStream *self)
 {
@@ -197,7 +213,9 @@ tf_stream_init (TfStream *self)
       TF_TYPE_STREAM, TfStreamPrivate);
 
   self->priv = priv;
+  g_static_mutex_init (&priv->mutex);
   priv->has_resource = TP_MEDIA_STREAM_DIRECTION_NONE;
+  priv->current_state = TP_MEDIA_STREAM_STATE_DISCONNECTED;
 }
 
 static void
@@ -338,6 +356,15 @@ tf_stream_dispose (GObject *object)
   TfStream *stream = TF_STREAM (object);
   TfStreamPrivate *priv = stream->priv;
 
+  TF_STREAM_LOCK (stream);
+  if (stream->priv->idle_connected_id)
+    g_source_remove (stream->priv->idle_connected_id);
+  stream->priv->idle_connected_id = 0;
+
+  stream->priv->disposed = TRUE;
+  TF_STREAM_UNLOCK (stream);
+
+
   if (priv->stream_handler_proxy)
     {
       TpMediaStreamHandler *tmp = priv->stream_handler_proxy;
@@ -368,6 +395,18 @@ tf_stream_dispose (GObject *object)
       priv->fs_session = NULL;
     }
 
+  if (priv->fs_participant)
+    {
+      g_object_unref (priv->fs_participant);
+      priv->fs_participant = NULL;
+    }
+
+  if (priv->fs_conference)
+    {
+      g_object_unref (priv->fs_conference);
+      priv->fs_conference = NULL;
+    }
+
   if (priv->local_preferences)
     {
       fs_codec_list_destroy (priv->local_preferences);
@@ -615,12 +654,14 @@ get_all_properties_cb (TpProxy *proxy,
   GPtrArray *dbus_relay_info = NULL;
   gboolean created_locally = TRUE;
   gboolean valid = FALSE;
+  guint i;
 
   if (dbus_error &&
       !(dbus_error->domain == DBUS_GERROR &&
           dbus_error->code == DBUS_GERROR_UNKNOWN_METHOD))
     {
-      tf_stream_error (stream, 0, dbus_error->message);
+      tf_stream_error (stream, TP_MEDIA_STREAM_ERROR_INVALID_CM_BEHAVIOR,
+          dbus_error->message);
       return;
     }
 
@@ -877,7 +918,7 @@ get_all_properties_cb (TpProxy *proxy,
 
   if (!stream->priv->fs_session)
     {
-      tf_stream_error (stream, 0, myerror->message);
+      tf_stream_error (stream, fserror_to_tperror (myerror), myerror->message);
       WARNING (stream, "Error creating session: %s", myerror->message);
       g_clear_error (&myerror);
       return;
@@ -894,9 +935,12 @@ get_all_properties_cb (TpProxy *proxy,
       params,
       &myerror);
 
+  for (i = 0; i < n_args; i++)
+    g_value_unset (&params[i].value);
+
   if (!stream->priv->fs_stream)
     {
-      tf_stream_error (stream, 0, myerror->message);
+      tf_stream_error (stream, fserror_to_tperror (myerror), myerror->message);
       WARNING (stream, "Error creating stream: %s", myerror->message);
       g_clear_error (&myerror);
       return;
@@ -910,7 +954,8 @@ get_all_properties_cb (TpProxy *proxy,
         if (!(myerror->domain == FS_ERROR &&
                 myerror->code == FS_ERROR_NOT_IMPLEMENTED))
           {
-            tf_stream_error (stream, 0, myerror->message);
+            tf_stream_error (stream, fserror_to_tperror (myerror),
+                myerror->message);
             WARNING (stream, "Error setting codec preferences: %s",
                 myerror->message);
             g_clear_error (&myerror);
@@ -980,7 +1025,7 @@ cb_fs_local_candidates_prepared (TfStream *self)
 
           g_value_init (&transport,
               TP_STRUCT_TYPE_MEDIA_STREAM_HANDLER_TRANSPORT);
-          g_value_set_static_boxed (&transport,
+          g_value_take_boxed (&transport,
               dbus_g_type_specialized_construct (
                   TP_STRUCT_TYPE_MEDIA_STREAM_HANDLER_TRANSPORT));
 
@@ -1029,7 +1074,7 @@ cb_fs_local_candidates_prepared (TfStream *self)
               9, candidate->password,
               G_MAXUINT);
 
-          g_ptr_array_add (transports, g_value_dup_boxed (&transport));
+          g_ptr_array_add (transports, g_value_get_boxed (&transport));
 
           self->priv->local_candidates = g_list_remove (
               self->priv->local_candidates, candidate);
@@ -1136,7 +1181,7 @@ tp_transports_to_fs (const gchar* foundation, const GPtrArray *transports)
           g_value_get_uint (g_value_array_get_nth (transport, 0)), /*component*/
           type, proto,
 
-          g_value_dup_string (g_value_array_get_nth (transport, 1)), /* ip */
+          g_value_get_string (g_value_array_get_nth (transport, 1)), /* ip */
           g_value_get_uint (g_value_array_get_nth (transport, 2))); /* port */
 
       fs_candidate->priority = (gint)
@@ -1213,7 +1258,7 @@ fs_codecs_to_tp (TfStream *stream,
         }
 
       g_value_init (&codec, TP_STRUCT_TYPE_MEDIA_STREAM_HANDLER_CODEC);
-      g_value_set_static_boxed (&codec,
+      g_value_take_boxed (&codec,
           dbus_g_type_specialized_construct (TP_STRUCT_TYPE_MEDIA_STREAM_HANDLER_CODEC));
 
       dbus_g_type_struct_set (&codec,
@@ -1252,7 +1297,7 @@ add_remote_candidate (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
 
   if (!fs_stream_set_remote_candidates (self->priv->fs_stream,
           fscandidates, &error))
-      tf_stream_error (self, 0, error->message);
+    tf_stream_error (self, fserror_to_tperror (error), error->message);
 
   fs_candidate_list_destroy (fscandidates);
   g_clear_error (&error);
@@ -1266,7 +1311,7 @@ remove_remote_candidate (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
 {
   TfStream *self = TF_STREAM (object);
 
-  tf_stream_error (self, 0,
+  tf_stream_error (self, TP_MEDIA_STREAM_ERROR_INVALID_CM_BEHAVIOR,
       "RemoveRemoteCandidate is DEPRECATED");
 }
 
@@ -1330,7 +1375,7 @@ set_remote_candidate_list (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
 
   if (!fs_stream_set_remote_candidates (self->priv->fs_stream,
                   fs_candidates, &error))
-    tf_stream_error (self, 0, error->message);
+    tf_stream_error (self, fserror_to_tperror (error), error->message);
 
   g_clear_error (&error);
   fs_candidate_list_destroy (fs_candidates);
@@ -1339,11 +1384,9 @@ set_remote_candidate_list (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
 static void
 fill_fs_params (gpointer key, gpointer value, gpointer user_data)
 {
-  GList **fs_params = (GList **) user_data;
-  FsCodecParameter *param = g_new0(FsCodecParameter,1);
-  param->name = key;
-  param->value = value;
-  *fs_params = g_list_prepend (*fs_params, param);
+  FsCodec *codec = user_data;
+
+  fs_codec_add_optional_parameter (codec, key, value);
 }
 
 static void
@@ -1355,7 +1398,6 @@ set_remote_codecs (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
   TfStream *self = TF_STREAM (object);
   GValueArray *codec;
   GHashTable *params = NULL;
-  GList *fs_params = NULL;
   GList *fs_remote_codecs = NULL;
   guint i;
   GError *error = NULL;
@@ -1388,10 +1430,7 @@ set_remote_codecs (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
           g_value_get_uint (g_value_array_get_nth (codec, 4));
 
       params = g_value_get_boxed (g_value_array_get_nth (codec, 5));
-      fs_params = NULL;
-      g_hash_table_foreach (params, fill_fs_params, &fs_params);
-
-      fs_codec->optional_params = fs_params;
+      g_hash_table_foreach (params, fill_fs_params, fs_codec);
 
       DEBUG (self, "adding remote codec %s [%d]",
           fs_codec->encoding_name, fs_codec->id);
@@ -1405,13 +1444,15 @@ set_remote_codecs (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
     /*
      * Call the error method with the proper thing here
      */
-    gchar *str = g_strdup_printf ("Codec negotiation failed: %s",
-        error->message);
-    tf_stream_error (self, 0, str);
-    g_free (str);
+    g_prefix_error (&error, "Codec negotiation failed: ");
+    tf_stream_error (self, fserror_to_tperror (error), error->message);
+    g_clear_error (&error);
+    fs_codec_list_destroy (fs_remote_codecs);
     return;
   }
 
+  fs_codec_list_destroy (fs_remote_codecs);
+
   self->priv->send_supported_codecs = TRUE;
   _tf_stream_try_sending_codecs (self);
 }
@@ -1450,7 +1491,8 @@ set_stream_playing (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
           }
         else
           {
-            tf_stream_error (self, 0, "Resource Unavailable");
+            tf_stream_error (self, TP_MEDIA_STREAM_ERROR_MEDIA_ERROR,
+                "Resource Unavailable");
           }
       }
       self->priv->desired_direction |= FS_DIRECTION_RECV;
@@ -1504,7 +1546,8 @@ set_stream_sending (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
           }
         else
           {
-            tf_stream_error (self, 0, "Resource Unavailable");
+            tf_stream_error (self, TP_MEDIA_STREAM_ERROR_MEDIA_ERROR,
+                "Resource Unavailable");
           }
       }
       self->priv->desired_direction |= FS_DIRECTION_SEND;
@@ -1627,7 +1670,8 @@ set_stream_held (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
         }
       else
         {
-          tf_stream_error (self, 0, "Error unholding stream");
+          tf_stream_error (self, TP_MEDIA_STREAM_ERROR_MEDIA_ERROR,
+              "Error unholding stream");
         }
     }
 }
@@ -1736,10 +1780,14 @@ cb_fs_new_active_candidate_pair (TfStream *self,
     async_method_callback, "Media.StreamHandler::NewActiveCandidatePair",
     NULL, (GObject *) self);
 
-  tp_cli_media_stream_handler_call_stream_state (
-    self->priv->stream_handler_proxy, -1, TP_MEDIA_STREAM_STATE_CONNECTED,
-    async_method_callback, "Media.StreamHandler::StreamState",
-    NULL, (GObject *) self);
+  if (self->priv->current_state == TP_MEDIA_STREAM_STATE_DISCONNECTED)
+  {
+    tp_cli_media_stream_handler_call_stream_state (
+        self->priv->stream_handler_proxy, -1, TP_MEDIA_STREAM_STATE_CONNECTED,
+        async_method_callback, "Media.StreamHandler::StreamState",
+        NULL, (GObject *) self);
+    self->priv->current_state = TP_MEDIA_STREAM_STATE_CONNECTED;
+  }
 }
 
 static void
@@ -1847,7 +1895,7 @@ _tf_stream_bus_message (TfStream *stream,
               enumvalue->value_nick, errorno, msg, debug);
           g_type_class_unref (enumclass);
 
-          tf_stream_error (stream, 0, msg);
+          tf_stream_error (stream, fserrorno_to_tperrorno (errorno), msg);
           return TRUE;
         }
     }
@@ -1961,6 +2009,48 @@ _tf_stream_bus_message (TfStream *stream,
           fs_codec_destroy (codec);
         }
     }
+  else if (gst_structure_has_name (s, "farsight-component-state-changed"))
+    {
+      FsStream *fsstream;
+      const GValue *value;
+      guint component;
+      FsStreamState fsstate;
+
+      value = gst_structure_get_value (s, "stream");
+      fsstream = g_value_get_object (value);
+
+      if (fsstream != stream->priv->fs_stream)
+        return FALSE;
+
+      if (!gst_structure_get_uint (s, "component", &component) ||
+          !gst_structure_get_enum (s, "state", FS_TYPE_STREAM_STATE,
+              (gint*) &fsstate))
+        return TRUE;
+
+      cb_fs_component_state_changed (stream, component, fsstate);
+    }
+
+  return FALSE;
+}
+
+static gboolean
+emit_connected (gpointer data)
+{
+  TfStream *self = TF_STREAM (data);
+
+  TF_STREAM_LOCK (self);
+  self->priv->idle_connected_id = 0;
+  if (self->priv->disposed)
+  {
+    TF_STREAM_UNLOCK (self);
+    return FALSE;
+  }
+  TF_STREAM_UNLOCK (self);
+
+  tp_cli_media_stream_handler_call_stream_state (
+      self->priv->stream_handler_proxy, -1, TP_MEDIA_STREAM_STATE_CONNECTED,
+      async_method_callback, "Media.StreamHandler::StreamState",
+      NULL, (GObject *) self);
 
   return FALSE;
 }
@@ -1977,6 +2067,17 @@ cb_fs_stream_src_pad_added (FsStream *fsstream G_GNUC_UNUSED,
   DEBUG (self, "New pad %s: " FS_CODEC_FORMAT, padname, FS_CODEC_ARGS (codec));
   g_free (padname);
 
+  TF_STREAM_LOCK (self);
+  if (self->priv->disposed)
+  {
+    TF_STREAM_UNLOCK (self);
+    return;
+  }
+
+  if (!self->priv->idle_connected_id)
+    self->priv->idle_connected_id = g_idle_add (emit_connected, self);
+  TF_STREAM_UNLOCK (self);
+
   g_signal_emit (self, signals[SRC_PAD_ADDED], 0, pad, codec);
 }
 
@@ -2103,3 +2204,81 @@ tf_stream_get_id (TfStream *stream)
 
   return stream->stream_id;
 }
+
+static TpMediaStreamError
+fserrorno_to_tperrorno (FsError fserror)
+{
+  TpMediaStreamError tperror;
+
+  switch (fserror)
+  {
+    case FS_ERROR_NETWORK:
+      tperror = TP_MEDIA_STREAM_ERROR_NETWORK_ERROR;
+      break;
+    case FS_ERROR_CONNECTION_FAILED:
+      tperror = TP_MEDIA_STREAM_ERROR_CONNECTION_FAILED;
+      break;
+    case FS_ERROR_NO_CODECS:
+      tperror = TP_MEDIA_STREAM_ERROR_NO_CODECS;
+      break;
+    case FS_ERROR_NEGOTIATION_FAILED:
+      tperror = TP_MEDIA_STREAM_ERROR_CODEC_NEGOTIATION_FAILED;
+      break;
+    case FS_ERROR_INVALID_ARGUMENTS:
+      tperror = TP_MEDIA_STREAM_ERROR_INVALID_CM_BEHAVIOR;
+      break;
+    case FS_ERROR_NO_CODECS_LEFT:
+    case FS_ERROR_CONSTRUCTION:
+    case FS_ERROR_INTERNAL:
+    case FS_ERROR_NOT_IMPLEMENTED: /* Not really a real error */
+    case FS_ERROR_DISPOSED: /* Not really a real error */
+    default:
+      tperror = TP_MEDIA_STREAM_ERROR_MEDIA_ERROR;
+  }
+
+  return tperror;
+}
+
+TpMediaStreamError
+fserror_to_tperror (GError *error)
+{
+  if (!error || error->domain != FS_ERROR)
+    return TP_MEDIA_STREAM_ERROR_UNKNOWN;
+
+  return fserrorno_to_tperrorno (error->code);
+}
+
+static void
+cb_fs_component_state_changed (TfStream *self,
+    guint component,
+    FsStreamState fsstate)
+{
+  TpMediaStreamState state;
+
+  if (component != 1)
+    return;
+
+  switch (fsstate)
+  {
+    case FS_STREAM_STATE_FAILED:
+    case FS_STREAM_STATE_DISCONNECTED:
+      state = TP_MEDIA_STREAM_STATE_DISCONNECTED;
+      break;
+    case FS_STREAM_STATE_GATHERING:
+    case FS_STREAM_STATE_CONNECTING:
+    case FS_STREAM_STATE_CONNECTED:
+      state = TP_MEDIA_STREAM_STATE_CONNECTING;
+      break;
+    case FS_STREAM_STATE_READY:
+    default:
+      state = TP_MEDIA_STREAM_STATE_CONNECTED;
+      break;
+  }
+
+  self->priv->current_state = state;
+
+  tp_cli_media_stream_handler_call_stream_state (
+      self->priv->stream_handler_proxy, -1, state,
+      async_method_callback, "Media.StreamHandler::StreamState",
+      NULL, (GObject *) self);
+}
